import logging
from functools import partial
from operator import is_not
from typing import Optional

from silx.utils.proxy import docstring
from nxtomo.io import HDF5File
from pyunitsystem.voltagesystem import VoltageSystem

from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.nxobject.nxdetector import NXdetector, NXdetectorWithUnit
from nxtomo.nxobject.nxsource import DefaultESRFSource, NXsource
from nxtomo.nxobject.nxobject import NXobject
from nxtomo.utils import get_data

_logger = logging.getLogger(__name__)


class NXinstrument(NXobject):
    def __init__(
        self, node_name: str = "instrument", parent: Optional[NXobject] = None
    ) -> None:
        super().__init__(node_name=node_name, parent=parent)
        self._set_freeze(False)
        self._detector = NXdetector(
            node_name="detector",
            parent=self,
            field_of_view="Full",
            expected_dim=(2, 3),
        )
        self._diode = NXdetectorWithUnit(
            node_name="diode",
            parent=self,
            expected_dim=(1,),
            default_unit=VoltageSystem.VOLT,
        )
        self._source = DefaultESRFSource(node_name="source", parent=self)
        self._name = None
        self._set_freeze(True)

    @property
    def detector(self) -> Optional[NXdetector]:
        return self._detector

    @detector.setter
    def detector(self, detector: Optional[NXdetector]):
        if not isinstance(detector, (NXdetector, type(None))):
            raise TypeError(
                f"detector is expected to be None or an instance of NXdetecetor. Not {type(detector)}"
            )
        self._detector = detector

    @property
    def diode(self) -> Optional[NXdetector]:
        return self._diode

    @diode.setter
    def diode(self, diode: Optional[NXdetector]):
        if not isinstance(diode, (NXdetector, type(None))):
            raise TypeError(
                f"diode is expected to be None or an instance of NXdetecetor. Not {type(diode)}"
            )
        self._diode = diode

    @property
    def source(self) -> Optional[NXsource]:
        return self._source

    @source.setter
    def source(self, source: Optional[NXsource]) -> None:
        if not isinstance(source, (NXsource, type(None))):
            raise TypeError(
                f"source is expected to be None or an instance of NXsource. Not {type(source)}"
            )
        self._source = source

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        if not isinstance(name, (str, type(None))):
            raise TypeError(
                f"name is expected to be None or an instance of str. Not {type(name)}"
            )
        self._name = name

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: Optional[float] = None,
        data_path: Optional[str] = None,
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_path_version)
        nexus_instrument_paths = nexus_paths.nx_instrument_paths
        nx_dict = {}

        if self._detector is not None:
            nx_dict.update(
                self._detector.to_nx_dict(nexus_path_version=nexus_path_version)
            )

        if self._diode is not None:
            nx_dict.update(
                self._diode.to_nx_dict(nexus_path_version=nexus_path_version)
            )

        if self._source is not None:
            nx_dict.update(
                self.source.to_nx_dict(nexus_path_version=nexus_path_version)
            )

        if self.name is not None:
            nx_dict[f"{self.name}/{nexus_instrument_paths.NAME}"] = self.name
        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXinstrument"

        return nx_dict

    def _load(
        self,
        file_path: str,
        data_path: str,
        nexus_version: float,
        detector_data_as: str,
    ) -> NXobject:
        """
        Create and load an NXsample from data on disk
        """
        nexus_paths = get_nexus_paths(nexus_version)
        nexus_instrument_paths = nexus_paths.nx_instrument_paths

        with HDF5File(file_path, mode="r") as h5f:
            if data_path in h5f:
                has_detector = "detector" in h5f[data_path]
                has_diode = "diode" in h5f[data_path]
                has_source = "source" in h5f[data_path]
            else:
                has_detector = False
                has_diode = False
                has_source = False
        # TODO: loading detector might be done using the NXclass instead of some hard coded names
        if has_detector:
            self.detector._load(
                file_path=file_path,
                data_path="/".join(
                    [data_path, "detector"],
                ),
                nexus_version=nexus_version,
                load_data_as=detector_data_as,
            )
        if has_diode:
            self.diode._load(
                file_path=file_path,
                data_path="/".join(
                    [data_path, "diode"],
                ),
                nexus_version=nexus_version,
                load_data_as="as_numpy_array",
            )
        if has_source:
            self.source._load(
                file_path=file_path,
                data_path="/".join([data_path, "source"]),
                nexus_version=nexus_version,
            )
        if nexus_instrument_paths.NAME is not None:
            self.name = get_data(
                file_path=file_path,
                data_path="/".join([data_path, nexus_instrument_paths.NAME]),
            )

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="instrument"):
        # filter None obj
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        if len(nx_objects) == 0:
            return None
        # warning: later we make the assumption that nx_objects contains at least one element
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXinstrument):
                raise TypeError("Cannot concatenate non NXinstrument object")

        nx_instrument = NXinstrument(node_name=node_name)

        nx_instrument.name = nx_objects[0].name
        _logger.info(f"instrument name {nx_objects[0].name} will be picked")

        nx_instrument.source = NXsource.concatenate(
            [nx_obj.source for nx_obj in nx_objects],
            node_name="source",
        )
        nx_instrument.source.parent = nx_instrument

        nx_instrument.diode = NXdetectorWithUnit.concatenate(
            [nx_obj.diode for nx_obj in nx_objects],
            node_name="diode",
            expected_dim=(1,),
            default_unit=VoltageSystem.VOLT,
        )
        nx_instrument.diode.parent = nx_instrument

        nx_instrument.detector = NXdetector.concatenate(
            [nx_obj.detector for nx_obj in nx_objects],
            node_name="detector",
        )
        nx_instrument.detector.parent = nx_instrument

        return nx_instrument



import logging
from functools import partial
from operator import is_not
from typing import Optional, Union

import numpy
from silx.utils.enum import Enum as _Enum
from silx.utils.proxy import docstring

from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.nxobject.nxobject import NXobject
from nxtomo.utils import get_data

_logger = logging.getLogger(__name__)


class SourceType(_Enum):
    SPALLATION_NEUTRON = "Spallation Neutron Source"
    PULSED_REACTOR_NEUTRON_SOURCE = "Pulsed Reactor Neutron Source"
    REACTOR_NEUTRON_SOURCE = "Reactor Neutron Source"
    SYNCHROTRON_X_RAY_SOURCE = "Synchrotron X-ray Source"
    PULSED_MUON_SOURCE = "Pulsed Muon Source"
    ROTATING_ANODE_X_RAY = "Rotating Anode X-ray"
    FIXED_TUBE_X_RAY = "Fixed Tube X-ray"
    UV_LASER = "UV Laser"
    FREE_ELECTRON_LASER = "Free-Electron Laser"
    OPTICAL_LASER = "Optical Laser"
    ION_SOURCE = "Ion Source"
    UV_PLASMA_SOURCE = "UV Plasma Source"
    METAL_JET_X_RAY = "Metal Jet X-ray"


class ProbeType(_Enum):
    NEUTRON = "neutron"
    X_RAY = "x-ray"
    MUON = "muon"
    ELECTRON = "electron"
    ULTRAVIOLET = "ultraviolet"
    VISIBLE_LIGHT = "visible light"
    POSITRON = "positron"
    PROTON = "proton"


class NXsource(NXobject):
    """Information regarding the x-ray storage ring/facility"""

    def __init__(
        self,
        node_name="source",
        parent=None,
        source_name=None,
        source_type=None,
        probe=None,
    ):
        super().__init__(node_name=node_name, parent=parent)
        self._set_freeze(False)
        self._name = source_name
        self._type = source_type
        self._probe = probe
        self._set_freeze(True)

    @property
    def name(self) -> Union[None, str]:
        return self._name

    @name.setter
    def name(self, source_name: Union[str, None]):
        if isinstance(source_name, numpy.ndarray):
            # handle Diamond Dataset
            source_name = source_name.tostring()
            if hasattr(source_name, "decode"):
                source_name = source_name.decode()
        if not isinstance(source_name, (str, type(None))):
            raise TypeError(
                f"source_name is expected to be None or a str not {type(source_name)}"
            )
        self._name = source_name

    @property
    def type(self) -> Optional[SourceType]:
        return self._type

    @type.setter
    def type(self, type_: Union[None, str, SourceType]):
        if type_ is None:
            self._type = None
        else:
            type_ = SourceType.from_value(type_)
            self._type = type_

    @property
    def probe(self) -> Optional[ProbeType]:
        return self._probe

    @probe.setter
    def probe(self, probe: Union[None, str, ProbeType]):
        if probe is None:
            self._probe = None
        else:
            self._probe = ProbeType.from_value(probe)

    def __str__(self):
        return f"{super().__str__}, (source name: {self.name}, source type: {self.type}, source probe: {self.probe})"

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: Optional[float] = None,
        data_path: Optional[str] = None,
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_path_version)
        nexus_source_paths = nexus_paths.nx_source_paths
        nx_dict = {}

        # warning: source is integrated only since 1.1 version of the nexus path
        if self.name is not None and nexus_paths.SOURCE_NAME is not None:
            path_name = f"{self.path}/{nexus_source_paths.NAME}"
            nx_dict[path_name] = self.name
        if self.type is not None and nexus_paths.SOURCE_TYPE is not None:
            path_type = f"{self.path}/{nexus_source_paths.TYPE}"
            nx_dict[path_type] = self.type.value
        if self.probe is not None and nexus_paths.SOURCE_PROBE is not None:
            path_probe = f"{self.path}/{nexus_source_paths.PROBE}"
            nx_dict[path_probe] = self.probe.value
        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXsource"

        return nx_dict

    def _load(self, file_path: str, data_path: str, nexus_version: float) -> None:
        nexus_paths = get_nexus_paths(nexus_version)
        nexus_source_paths = nexus_paths.nx_source_paths
        self.name = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_source_paths.NAME]),
        )
        try:
            self.type = get_data(
                file_path=file_path,
                data_path="/".join([data_path, nexus_source_paths.TYPE]),
            )
        except ValueError as e:
            _logger.warning(f"Fail to load source type. Error is {e}")
        try:
            self.probe = get_data(
                file_path=file_path,
                data_path="/".join([data_path, nexus_source_paths.PROBE]),
            )
        except ValueError as e:
            _logger.warning(f"Fail to load probe. Error is {e}")

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="source"):
        # filter None obj
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        if len(nx_objects) == 0:
            return None
        # warning: later we make the assumption that nx_objects contains at least one element
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXsource):
                raise TypeError("Cannot concatenate non NXsource object")

        nx_souce = NXsource(node_name=node_name)
        nx_souce.name = nx_objects[0].name
        _logger.info(f"Take the first source name {nx_objects[0].name}")
        nx_souce.type = nx_objects[0].type
        _logger.info(f"Take the first source type {nx_objects[0].type}")
        nx_souce.probe = nx_objects[0].probe
        _logger.info(f"Take the first source probe {nx_objects[0].probe}")
        return nx_souce


class DefaultESRFSource(NXsource):
    def __init__(self, node_name="source", parent=None) -> None:
        super().__init__(
            node_name=node_name,
            parent=parent,
            source_name="ESRF",
            source_type=SourceType.SYNCHROTRON_X_RAY_SOURCE,
            probe=ProbeType.X_RAY,
        )



from typing import Iterable

import h5py
from silx.io.utils import h5py_read_dataset

from nxtomo.io import HDF5File
from nxtomo.nxobject.nxobject import NXobject


def create_nx_data_group(file_path: str, entry_path: str, axis_scale: list):
    """
    Create the 'Nxdata' group at entry level with soft links on the NXDetector
    and NXsample.

    :param file_path:
    :param entry_path:
    :param axis_scale:
    :return:
    """
    if not isinstance(file_path, str):
        raise TypeError("file_path is expected to be a file")
    if not isinstance(entry_path, str):
        raise TypeError("entry_path is expected to be a file")
    if not isinstance(axis_scale, Iterable):
        raise TypeError("axis_scale is expected to be an Iterable")

    with HDF5File(filename=file_path, mode="a") as h5f:
        entry_group = h5f[entry_path]

        nx_data_grp = entry_group.require_group("data")
        # link detector datasets:
        if not entry_path.startswith("/"):
            entry_path = "/" + entry_path
        for dataset in ("data", "image_key", "image_key_control"):
            dataset_path = "/".join((entry_path, "instrument", "detector", dataset))
            nx_data_grp[dataset] = h5py.SoftLink(dataset_path)
        # link rotation angle
        nx_data_grp["rotation_angle"] = h5py.SoftLink(
            "/".join((entry_path, "sample", "rotation_angle"))
        )

        # write NX attributes
        nx_data_grp.attrs["NX_class"] = "NXdata"
        nx_data_grp.attrs["signal"] = "data"
        nx_data_grp.attrs["SILX_style/axis_scale_types"] = axis_scale
        nx_data_grp["data"].attrs["interpretation"] = "image"


def link_nxbeam_to_root(file_path, entry_path):
    """
    Create the 'Nxdata' group at entry level with soft links on the NXDetector
    and NXsample.

    :param file_path:
    :param entry_path:
    :return:
    """
    if not isinstance(file_path, str):
        raise TypeError("file_path is expected to be a file")
    if not isinstance(entry_path, str):
        raise TypeError("entry_path is expected to be a file")

    if not entry_path.startswith("/"):
        entry_path = "/" + entry_path
    with HDF5File(filename=file_path, mode="a") as h5f:
        entry_group = h5f[entry_path]
        entry_group["beam"] = h5py.SoftLink(
            "/".join((entry_path, "instrument", "beam"))
        )


def get_data_and_unit(file_path, data_path, default_unit):
    with HDF5File(file_path, mode="r") as h5f:
        if data_path in h5f and isinstance(h5f[data_path], h5py.Dataset):
            dataset = h5f[data_path]
            unit = None
            if "unit" in dataset.attrs:
                unit = dataset.attrs["unit"]
            elif "units" in dataset.attrs:
                unit = dataset.attrs["units"]
            else:
                unit = default_unit
            if hasattr(unit, "decode"):
                # handle Diamond dataset
                unit = unit.decode()
            return h5py_read_dataset(dataset), unit
        else:
            return None, default_unit


def get_data(file_path, data_path):
    with HDF5File(file_path, mode="r") as h5f:
        if data_path in h5f:
            return h5py_read_dataset(h5f[data_path])
        else:
            return None


def concatenate(nx_objects: Iterable, **kwargs):
    if len(nx_objects) == 0:
        return None
    else:
        if not isinstance(nx_objects[0], NXobject):
            raise TypeError("nx_objects are expected to be instances of NXobject")
        return type(nx_objects[0]).concatenate(nx_objects=nx_objects, **kwargs)



"""module to provide helper classes to define transformations contained in NXtransformations"""

import typing
import logging
import numpy

from pyunitsystem.metricsystem import MetricSystem

from silx.utils.enum import Enum as _Enum


_logger = logging.getLogger(__name__)


class TransformationType(_Enum):
    TRANSLATION = "translation"
    ROTATION = "rotation"


class TransformationAxis:
    """
    Some predefined axis for tomography acquisition done at esrf.
    Warning those are stored as (X, Y, Z) and not under the usual numpy reference (Z, Y, X)

    space is defined here: https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html
    """

    AXIS_X = (1, 0, 0)
    AXIS_Y = (0, 1, 0)
    AXIS_Z = (0, 0, 1)


class Transformation:
    """
    Define a Transformation done on an axis

    :param str axis_name: name of the transformation.
    :param TransformationType transformation_type: type of the formation. As unit depends on the type of transformation this is not possible to modify it once created
    :param vector: vector of the transformation. Expected as a tuple of three values that define the axis for this transformation. Can also be an instance of TransformationAxis predefining some default axis
    :param Optional[str] depends_on: used to determine transformation chain. If depends on no other transformation then should be considered as if it is depending on "gravity" only.
    :warning: when convert a rotation which as 'radian' as units it will be cast to degree
    """

    __isfrozen = False
    # to ease API and avoid setting wrong attributes we 'freeze' the attributes
    # see https://stackoverflow.com/questions/3603502/prevent-creating-new-attributes-outside-init

    def __init__(
        self,
        axis_name: str,
        value,
        transformation_type: TransformationType,
        vector: typing.Union[typing.Tuple[float, float, float], TransformationAxis],
        depends_on: typing.Optional[str] = None,
    ) -> None:
        self._axis_name = axis_name
        self._transformation_values = None
        self.transformation_values = value
        self._transformation_type = TransformationType.from_value(transformation_type)
        self._units = (
            MetricSystem.METER
            if self._transformation_type is TransformationType.TRANSLATION
            else "degree"
        )
        if isinstance(vector, TransformationAxis):
            self._vector = vector.value()
        elif not isinstance(vector, (tuple, list, numpy.ndarray)) or len(vector) != 3:
            raise TypeError(
                f"vector should be a tuple of three elements. {vector} provided"
            )
        else:
            self._vector = tuple(vector)
            assert len(self._vector) == 3, ""
        self._offset = (0, 0, 0)
        self._depends_on = None
        self.depends_on = depends_on
        self._equipment_component = None
        self._set_freeze()

    def _set_freeze(self, freeze=True):
        self.__isfrozen = freeze

    @property
    def axis_name(self) -> str:
        return self._axis_name

    @axis_name.setter
    def axis_name(self, axis_name: str):
        self._axis_name = axis_name

    @property
    def transformation_values(self):
        return self._transformation_values

    @transformation_values.setter
    def transformation_values(self, values):
        if values is not None and not numpy.isscalar(values):
            self._transformation_values = numpy.array(values)
        else:
            self._transformation_values = values

    @property
    def transformation_type(self) -> TransformationType:
        return self._transformation_type

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, units: typing.Union[str, MetricSystem]):
        """
        :raises ValueError: if units is invalid (depends on the transformation type).
        """
        if units == "m/s2":
            # corner cases of the gravity
            self._units = units
        elif self._transformation_type is TransformationType.TRANSLATION:
            self._units = MetricSystem.from_value(units)
        elif units in ("rad", "radian", "radians"):
            self._units = "rad"
        elif units in ("degree", "degrees"):
            self._units = "degree"
        else:
            raise ValueError(f"Unrecognized unit {units}")

    @property
    def vector(self) -> typing.Tuple[float, float, float]:
        return self._vector

    @property
    def offset(self) -> tuple:
        return self._offset

    @offset.setter
    def offset(self, offset: typing.Union[tuple, list, numpy.ndarray]):
        if not isinstance(offset, (tuple, list, numpy.ndarray)):
            raise TypeError(
                f"offset is expected to be a vector of three elements. {type(offset)} provided"
            )
        elif not len(offset) == 3:
            raise TypeError(
                f"offset is expected to be a vector of three elements. {offset} provided"
            )
        self._offset = tuple(offset)

    @property
    def depends_on(self):
        return self._depends_on

    @depends_on.setter
    def depends_on(self, depends_on):
        """
        :param Optional[Transformation] depends_on:
        """
        if not (depends_on is None or isinstance(depends_on, str)):
            raise TypeError(
                f"offset is expected to be None or str. {type(depends_on)} provided"
            )
        self._depends_on = depends_on

    @property
    def equipment_component(self) -> typing.Optional[str]:
        return self._equipment_component

    @equipment_component.setter
    def equipment_component(self, equipment_component: typing.Optional[str]):
        if not (equipment_component is None or isinstance(equipment_component, str)):
            raise TypeError(
                f"equipment_component is expect to ne None or a str. {type(equipment_component)} provided"
            )
        self._equipment_component = equipment_component

    def get_transformation_values_in_common_unit(self):
        transformation_values = self.transformation_values
        units = self.units
        if units in ("radian", "rad", "rads", "radians"):
            if transformation_values is None:
                return None, "degree"
            else:
                transformation_values = numpy.rad2deg(transformation_values)
                return transformation_values % 360, "degree"
        elif units in ("degree", "degrees"):
            if transformation_values is None:
                return None, "degree"
            else:
                return transformation_values % 360, "degree"
        else:
            converted_values = (
                transformation_values * MetricSystem.from_str(str(units)).value
            )
            return converted_values, MetricSystem.METER

    def to_nx_dict(self, transformations_nexus_paths, data_path: str):
        def join(my_list):
            # filter empty strings
            my_list = tuple(
                filter(
                    lambda a: bool(
                        a
                    ),  # return False if a is an empty string else True,
                    my_list,
                )
            )
            if len(my_list) == 0:
                return ""
            else:
                return "/".join(my_list)

        transformation_values = self.transformation_values
        if transformation_values is None:
            _logger.error(f"no values defined for {self.axis_name}")
        elif numpy.isscalar(transformation_values):
            pass
        else:
            transformation_values = numpy.array(transformation_values)
        units = self.units
        if units == "radian":
            if transformation_values is not None:
                transformation_values = numpy.rad2deg(transformation_values)
            units = "degree"
        elif isinstance(units, MetricSystem):
            units = str(units)

        res = {
            join((data_path, self.axis_name)): self.transformation_values,
            join(
                (
                    data_path,
                    self.axis_name + transformations_nexus_paths.TRANSFORMATION_TYPE,
                )
            ): self.transformation_type.value,
            join((data_path, f"{self.axis_name}@units")): units,
        }

        # vector is mandatory
        res[
            join((data_path, f"{self.axis_name}{transformations_nexus_paths.VECTOR}"))
        ] = self.vector
        if self.offset is not None:
            res[
                join(
                    (data_path, f"{self.axis_name}{transformations_nexus_paths.OFFSET}")
                )
            ] = self.offset
        if self.depends_on:
            res[
                join(
                    (
                        data_path,
                        f"{self.axis_name}{transformations_nexus_paths.DEPENDS_ON}",
                    )
                )
            ] = self.depends_on
        if self.equipment_component:
            res[
                join(
                    (
                        data_path,
                        f"{self.axis_name}{transformations_nexus_paths.EQUIPMENT_COMPONENT}",
                    )
                )
            ] = self.equipment_component
        return res

    @staticmethod
    def from_nx_dict(axis_name: str, dict_: dict, transformations_nexus_paths):
        if transformations_nexus_paths is None:
            _logger.warning(
                "no transformations_nexus_paths (not implemented on this version of nexus - too old)"
            )
            return None
        value = dict_.get(axis_name, None)
        vector = dict_.get(f"{axis_name}{transformations_nexus_paths.VECTOR}", None)
        transformation_type = dict_.get(
            f"{axis_name}{transformations_nexus_paths.TRANSFORMATION_TYPE}", None
        )
        if vector is None or transformation_type is None:
            raise ValueError(
                "Unable to find mandatory vector and/or transformation_type"
            )

        transformation = Transformation(
            axis_name=axis_name,
            value=value,
            transformation_type=transformation_type,
            vector=vector,
        )

        units = dict_.get(f"{axis_name}@units", None) or dict_.get(
            "{axis_name}@unit", None
        )
        if units is not None:
            transformation.units = units

        offset = dict_.get(f"{axis_name}{transformations_nexus_paths.OFFSET}", None)
        if offset is not None:
            transformation.offset = offset

        depends_on = dict_.get(
            f"{axis_name}{transformations_nexus_paths.DEPENDS_ON}", None
        )
        if depends_on is not None:
            transformation.depends_on = depends_on

        equipment_component = dict_.get(
            f"{axis_name}{transformations_nexus_paths.EQUIPMENT_COMPONENT}", None
        )
        if equipment_component is not None:
            transformation.equipment_component = equipment_component

        return transformation

    def __setattr__(self, __name, __value):
        if self.__isfrozen and not hasattr(self, __name):
            raise AttributeError("can't set attribute", __name)
        else:
            super().__setattr__(__name, __value)

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Transformation):
            return False
        else:
            if not (
                self.vector == __value.vector
                and self.transformation_type == __value.transformation_type
                and self.offset == __value.offset
                and self.depends_on == __value.depends_on
                and self.equipment_component == __value.equipment_component
            ):
                return False
            else:
                values_a, units_a = self.get_transformation_values_in_common_unit()
                values_b, units_b = __value.get_transformation_values_in_common_unit()
                if values_a is None or values_b is None:
                    return (values_a is values_b) and (units_a == units_b)
                elif units_a != units_b:
                    return False
                if isinstance(values_a, numpy.ndarray) and isinstance(
                    values_b, numpy.ndarray
                ):
                    return numpy.array_equal(values_a, values_b)
                else:
                    return values_a == values_b

    def as_matrix(self):
        # handle the transformation
        if self.transformation_values is None:
            raise ValueError(f"missing transformation values for {self}")
        elif numpy.isscalar(self.transformation_values):
            if self.transformation_type is TransformationType.ROTATION:
                if self.units in ("rad", "rads", "radian", "radians"):
                    theta = self.transformation_values
                elif self.units in ("deg", "degree", "degs", "degrees"):
                    theta = numpy.deg2rad(self.transformation_values)
                else:
                    raise ValueError(f"unknow unit: {self.units}")

                if self.offset != (0, 0, 0):
                    raise ValueError("offset not handled")

                if self.vector == (1, 0, 0):
                    return numpy.array(
                        [
                            [
                                1,
                                0,
                                0,
                            ],
                            [0, numpy.cos(theta), -numpy.sin(theta)],
                            [0, numpy.sin(theta), -numpy.cos(theta)],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 1, 0):
                    return numpy.array(
                        [
                            [numpy.cos(theta), 0, numpy.sin(theta)],
                            [0, 1, 0],
                            [-numpy.sin(theta), 0, numpy.cos(theta)],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 0, 1):
                    return numpy.array(
                        [
                            [numpy.cos(theta), -numpy.sin(theta), 0],
                            [numpy.sin(theta), numpy.cos(theta), 0],
                            [0, 0, 1],
                        ],
                        dtype=numpy.float32,
                    )
                else:
                    raise ValueError(f"vector {self.vector} not handled")
            elif self.transformation_type is TransformationType.TRANSLATION:
                if self.vector == (1, 0, 0):
                    return numpy.array(
                        [
                            [
                                self.transformation_values,
                                0,
                                0,
                            ],
                            [0, 1, 0],
                            [0, 0, 1],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 1, 0):
                    return numpy.array(
                        [
                            [1, 0, 0],
                            [0, self.transformation_values, 0],
                            [0, 0, 1],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 0, 1):
                    return numpy.array(
                        [
                            [1, 0, 0],
                            [0, 1, 0],
                            [0, 0, self.transformation_values],
                        ],
                        dtype=numpy.float32,
                    )
            else:
                raise RuntimeError(
                    f"unknow transformation type: {self.transformation_type}"
                )
        else:
            raise ValueError(
                f"transformations as a list of values is not handled for now ({self})"
            )


class UDDetTransformation(Transformation):
    """
    convenient class to define a detector up-down flip if we consider the center of the detector to be at (0, 0)
    """

    def __init__(
        self,
        axis_name="ry",
        depends_on=None,
    ) -> None:
        super().__init__(
            axis_name=axis_name,
            value=180,
            transformation_type=TransformationType.ROTATION,
            vector=TransformationAxis.AXIS_Y,
            depends_on=depends_on,
        )


class LRDetTransformation(Transformation):
    """
    convenient class to define a detector left-right flip if we consider the center of the detector to be at (0, 0)
    """

    def __init__(
        self,
        axis_name="rz",
        depends_on=None,
    ) -> None:
        super().__init__(
            axis_name=axis_name,
            value=180,
            transformation_type=TransformationType.ROTATION,
            vector=TransformationAxis.AXIS_Z,
            depends_on=depends_on,
        )


class GravityTransformation(Transformation):
    """
    Gravity is used to solve transformation chain (as chain 'endpoint')
    """

    def __init__(self) -> None:
        super().__init__(
            axis_name="gravity",
            value=numpy.nan,
            transformation_type=TransformationType.TRANSLATION,
            vector=(0, 0, -1),
        )
        self.units = "m/s2"


def get_lr_flip(transformations: tuple) -> tuple:
    """
    check along all transformations if find Transformation matching 'LRTransformation'

    return a tuple with all matching keys
    """
    if not isinstance(transformations, (tuple, list)):
        raise TypeError(
            f"transformations is expected to be a tuple. {type(transformations)} provided"
        )
    res = []
    for transformation in transformations:
        if transformation == LRDetTransformation():
            res.append(transformation)
    return tuple(res)


def get_ud_flip(transformations: tuple) -> tuple:
    """
    check along all transformations if find Transformation matching 'UDTransformation'

    return a tuple with all matching keys
    """
    if not isinstance(transformations, (tuple, list)):
        raise TypeError(
            f"transformations is expected to be a tuple. {type(transformations)} provided"
        )
    res = []
    for transformation in transformations:
        if transformation == UDDetTransformation():
            res.append(transformation)
    return tuple(res)


def build_matrix(transformations: set):
    """
    build a matrice from a set of Transformation
    """
    transformations = {
        transformation.axis_name: transformation for transformation in transformations
    }
    already_applied_transformations = set(["gravity"])

    def handle_transformation(transformation: Transformation, matrix):
        if not isinstance(transformation, Transformation):
            raise TypeError(
                f"transformation is expected to be an instance of {Transformation}. {type(transformation)} provided"
            )

        # handle dependancies
        if transformation.axis_name in already_applied_transformations:
            # case already applied
            return matrix
        elif transformation.transformation_values is None:
            # case of the gravity matrix
            if transformation.axis_name.lower() == "gravity":
                return numpy.identity(3, dtype=numpy.float32)
            else:
                _logger.error(
                    f"transformation value not provided for {transformation.axis_name}. Ignore the transformation"
                )
                return matrix
        elif transformation.depends_on not in already_applied_transformations:
            if transformation.depends_on not in transformations:
                raise ValueError(
                    f"Unable to find transformation {transformation.depends_on}. Unable to build matrix. reason is: broken dependancy chain"
                )
            else:
                matrix = handle_transformation(
                    transformations[transformation.depends_on], matrix
                )
        print("update matrice")
        matrix = numpy.matmul(matrix, transformation.as_matrix())
        already_applied_transformations.add(transformation.axis_name)
        return matrix

    matrix = numpy.identity(3, dtype=numpy.float32)
    for transformation in transformations.values():
        matrix = handle_transformation(transformation, matrix)

    return matrix



import logging
from typing import Optional

from silx.utils.deprecation import deprecated

import nxtomo
from nxtomo.paths import (
    nxdetector,
    nxinstrument,
    nxmonitor,
    nxsample,
    nxsource,
    nxtransformations,
)

_logger = logging.getLogger(__name__)


LATEST_VERSION = 1.3


class NXtomo_PATH:
    # list all path that can be used by an nxtomo entry and read by nxtomo.
    # this is also used by nxtomomill to know were to save data

    _NX_DETECTOR_PATHS = None
    _NX_INSTRUMENT_PATHS = None
    _NX_SAMPLE_PATHS = None
    _NX_SOURCE_PATHS = None
    _NX_CONTROL_PATHS = None

    _NX_TRANSFORMATIONS_PATHS = None
    # paths used per each transformation contained in NX_TRANSFORMATIONS

    VERSION = None

    @property
    def nx_detector_paths(self):
        return self._NX_DETECTOR_PATHS

    @property
    def nx_instrument_paths(self):
        return self._NX_INSTRUMENT_PATHS

    @property
    def nx_sample_paths(self):
        return self._NX_SAMPLE_PATHS

    @property
    def nx_source_paths(self):
        return self._NX_SOURCE_PATHS

    @property
    def nx_monitor_paths(self):
        return self._NX_CONTROL_PATHS

    @property
    def nx_transformations_paths(self):
        return self._NX_TRANSFORMATIONS_PATHS

    @property
    def PROJ_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.DATA,
            ]
        )

    @property
    def SCAN_META_PATH(self) -> str:
        # for now scan_meta and technique are not link to any nxtomo...
        return "scan_meta/technique/scan"

    @property
    def INSTRUMENT_PATH(self) -> str:
        return "instrument"

    @property
    def CONTROL_PATH(self) -> str:
        return "control"

    @property
    def DET_META_PATH(self) -> str:
        return "scan_meta/technique/detector"

    @property
    def ROTATION_ANGLE_PATH(self):
        return "/".join(["sample", self.nx_sample_paths.ROTATION_ANGLE])

    @property
    def SAMPLE_PATH(self) -> str:
        return "sample"

    @property
    def NAME_PATH(self) -> str:
        return "sample/name"

    @property
    def GRP_SIZE_ATTR(self) -> str:
        return "group_size"

    @property
    def SAMPLE_NAME_PATH(self) -> str:
        return "/".join([self.SAMPLE_PATH, self.nx_sample_paths.NAME])

    @property
    def X_TRANS_PATH(self) -> str:
        return "/".join([self.SAMPLE_PATH, self.nx_sample_paths.X_TRANSLATION])

    @property
    def Y_TRANS_PATH(self) -> str:
        return "/".join([self.SAMPLE_PATH, self.nx_sample_paths.Y_TRANSLATION])

    @property
    def Z_TRANS_PATH(self) -> str:
        return "/".join([self.SAMPLE_PATH, self.nx_sample_paths.Z_TRANSLATION])

    @property
    def IMG_KEY_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.IMAGE_KEY,
            ]
        )

    @property
    def IMG_KEY_CONTROL_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.IMAGE_KEY_CONTROL,
            ]
        )

    @property
    def X_PIXEL_SIZE_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.X_PIXEL_SIZE,
            ]
        )

    @property
    def Y_PIXEL_SIZE_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.Y_PIXEL_SIZE,
            ]
        )

    @property
    def X_REAL_PIXEL_SIZE_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.X_REAL_PIXEL_SIZE,
            ]
        )

    @property
    def Y_REAL_PIXEL_SIZE_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.Y_REAL_PIXEL_SIZE,
            ]
        )

    @property
    @deprecated(replacement="X_PIXEL_SIZE_PATH", since_version="1.1.0")
    def X_PIXEL_MAG_SIZE_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.X_PIXEL_SIZE_MAGNIFIED,
            ]
        )

    @property
    @deprecated(replacement="Y_PIXEL_SIZE_PATH", since_version="1.1.0")
    def Y_PIXEL_MAG_SIZE_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.Y_PIXEL_SIZE_MAGNIFIED,
            ]
        )

    @property
    def DISTANCE_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.DISTANCE,
            ]
        )

    @property
    def FOV_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.FOV,
            ]
        )

    @property
    def EXPOSURE_TIME_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.EXPOSURE_TIME,
            ]
        )

    @property
    def ELECTRIC_CURRENT_PATH(self) -> str:
        return "/".join(
            [
                self.CONTROL_PATH,
                self.nx_monitor_paths.DATA_PATH,
            ]
        )

    @property
    def ESTIMATED_COR_FRM_MOTOR_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DETECTOR_PATH,
                self.nx_detector_paths.ESTIMATED_COR_FRM_MOTOR,
            ]
        )

    @property
    def TOMO_N_SCAN(self) -> str:
        return "/".join(
            [self.INSTRUMENT_PATH, self.nx_instrument_paths.DETECTOR_PATH, "tomo_n"]
        )

    @property
    def BEAM_PATH(self) -> str:
        return "beam"

    @property
    def ENERGY_PATH(self) -> str:
        return f"{self.BEAM_PATH}/incident_energy"

    @property
    def START_TIME_PATH(self) -> str:
        return "start_time"

    @property
    def END_TIME_PATH(self) -> str:
        return "end_time"

    @property
    @deprecated(replacement="END_TIME_PATH", reason="typo", since_version="0.8.0")
    def END_TIME_START(self) -> str:
        return self.END_TIME_PATH

    @property
    def INTENSITY_MONITOR_PATH(self) -> str:
        return "diode/data"

    @property
    @deprecated(
        replacement="", reason="will be removed. Not used", since_version="0.8.0"
    )
    def EPSILON_ROT_ANGLE(self) -> float:
        return 0.02

    @property
    def SOURCE_NAME(self) -> Optional[str]:
        return None

    @property
    def SOURCE_TYPE(self) -> Optional[str]:
        return None

    @property
    def SOURCE_PROBE(self) -> Optional[str]:
        return None

    @property
    def INSTRUMENT_NAME(self) -> Optional[str]:
        return None

    @property
    def ROCKING_PATH(self) -> str:
        return "/".join([self.SAMPLE_PATH, self.nx_sample_paths.ROCKING])

    @property
    def BASE_TILT_PATH(self) -> str:
        return "/".join([self.SAMPLE_PATH, self.nx_sample_paths.BASE_TILT])


class NXtomo_PATH_v_1_0(NXtomo_PATH):
    VERSION = 1.0

    _NX_DETECTOR_PATHS = nxdetector.NEXUS_DETECTOR_PATH_V_1_0
    _NX_INSTRUMENT_PATHS = nxinstrument.NEXUS_INSTRUMENT_PATH_V_1_0
    _NX_SAMPLE_PATHS = nxsample.NEXUS_SAMPLE_PATH_V_1_0
    _NX_SOURCE_PATHS = nxsource.NEXUS_SOURCE_PATH_V_1_0
    _NX_CONTROL_PATHS = nxmonitor.NEXUS_MONITOR_PATH_V_1_1


nx_tomo_path_v_1_0 = NXtomo_PATH_v_1_0()


class NXtomo_PATH_v_1_1(NXtomo_PATH_v_1_0):
    VERSION = 1.1

    _NX_DETECTOR_PATHS = nxdetector.NEXUS_DETECTOR_PATH_V_1_1
    _NX_INSTRUMENT_PATHS = nxinstrument.NEXUS_INSTRUMENT_PATH_V_1_1
    _NX_SAMPLE_PATHS = nxsample.NEXUS_SAMPLE_PATH_V_1_1
    _NX_SOURCE_PATHS = nxsource.NEXUS_SOURCE_PATH_V_1_1

    @property
    def NAME_PATH(self) -> str:
        return "title"

    @property
    def BEAM_PATH(self) -> str:
        return "/".join([self.INSTRUMENT_PATH, self.nx_instrument_paths.BEAM])

    @property
    def SOURCE_NAME(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.SOURCE,
                self.nx_source_paths.NAME,
            ]
        )

    @property
    def SOURCE_TYPE(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.SOURCE,
                self.nx_source_paths.TYPE,
            ]
        )

    @property
    def SOURCE_PROBE(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.SOURCE,
                self.nx_source_paths.PROBE,
            ]
        )

    @property
    def INSTRUMENT_NAME(self) -> str:
        return "/".join([self.INSTRUMENT_PATH, self.nx_instrument_paths.NAME])


nx_tomo_path_v_1_1 = NXtomo_PATH_v_1_1()


class NXtomo_PATH_v_1_2(NXtomo_PATH_v_1_1):
    VERSION = 1.2

    _NX_DETECTOR_PATHS = nxdetector.NEXUS_DETECTOR_PATH_V_1_2
    _NX_INSTRUMENT_PATHS = nxinstrument.NEXUS_INSTRUMENT_PATH_V_1_2
    _NX_SAMPLE_PATHS = nxsample.NEXUS_SAMPLE_PATH_V_1_2
    _NX_SOURCE_PATHS = nxsource.NEXUS_SOURCE_PATH_V_1_2

    @property
    def INTENSITY_MONITOR_PATH(self) -> str:
        return "/".join(
            [
                self.INSTRUMENT_PATH,
                self.nx_instrument_paths.DIODE,
                self.nx_detector_paths.DATA,
            ]
        )


nx_tomo_path_v_1_2 = NXtomo_PATH_v_1_2()


class NXtomo_PATH_v_1_3(NXtomo_PATH_v_1_2):
    VERSION = 1.3
    _NX_DETECTOR_PATHS = nxdetector.NEXUS_DETECTOR_PATH_V_1_3
    _NX_INSTRUMENT_PATHS = nxinstrument.NEXUS_INSTRUMENT_PATH_V_1_3
    _NX_SAMPLE_PATHS = nxsample.NEXUS_SAMPLE_PATH_V_1_3
    _NX_SOURCE_PATHS = nxsource.NEXUS_SOURCE_PATH_V_1_3
    _NX_TRANSFORMATIONS_PATHS = nxtransformations.NEXUS_TRANSFORMATIONS_PATH_V_1_3


nx_tomo_path_v_1_3 = NXtomo_PATH_v_1_3()


nx_tomo_path_latest = nx_tomo_path_v_1_3


def get_paths(version: Optional[float]) -> NXtomo_PATH:
    if version is None:
        version = LATEST_VERSION
        _logger.warning(
            f"version of the NXtomo not found. Will take the latest one ({LATEST_VERSION})"
        )
    versions_dict = {
        # Ensure compatibility with "old" datasets (acquired before Dec. 2021).
        # nxtomo can still parse them provided that nx_version=1.0 is forced at init.
        0.0: nx_tomo_path_v_1_0,
        0.1: nx_tomo_path_v_1_0,
        #
        1.0: nx_tomo_path_v_1_0,
        1.1: nx_tomo_path_v_1_1,
        1.2: nx_tomo_path_v_1_2,
        1.3: nx_tomo_path_v_1_3,
    }
    if version not in versions_dict:
        if int(version) == 1:
            _logger.warning(
                f"nexus path {version} requested but unknow from this version of nxtomo {nxtomo.__version__}. Pick latest one of this major version. You might miss some information"
            )
            version = LATEST_VERSION
        else:
            raise ValueError(f"Unknow major version of the nexus path ({version})")
    return versions_dict[version]



import os
import tempfile

import h5py
import numpy.random
import pytest
from silx.io.url import DataUrl
from nxtomo.io import cwd_context
from pyunitsystem.voltagesystem import VoltageSystem
from nxtomo.utils.transformation import TransformationAxis

from nxtomo.nxobject.nxtransformations import Transformation
from nxtomo.nxobject.nxdetector import (
    FieldOfView,
    NXdetector,
    NXdetectorWithUnit,
    ImageKey,
)


def test_nx_detector():
    """test creation and saving of an nxdetector"""
    nx_detector = NXdetector(expected_dim=(2, 3))

    # check data
    with pytest.raises(TypeError):
        nx_detector.data = 12
    # if expected dims is not fulfill
    with pytest.raises(ValueError):
        nx_detector.data = numpy.random.random(100 * 100 * 5).reshape(5, 10, 10, 100)
    with pytest.raises(TypeError):
        nx_detector.data = (
            12,
            13,
        )
    nx_detector.data = numpy.random.random(100 * 100 * 5).reshape(5, 100, 100)

    # check image key control
    with pytest.raises(TypeError):
        nx_detector.image_key_control = 12
    nx_detector.image_key_control = [1] * 5
    nx_detector.image_key_control = [ImageKey.PROJECTION] * 5

    # check x and y pixel size (both 'real' and 'sample')
    with pytest.raises(TypeError):
        nx_detector.x_pixel_size = "test"
    nx_detector.x_pixel_size = 1e-7

    with pytest.raises(TypeError):
        nx_detector.y_pixel_size = {}
    nx_detector.y_pixel_size = 2e-7

    # check x_flipped and y_flipped
    with pytest.raises(TypeError):
        nx_detector.x_flipped = 12
    with pytest.raises(TypeError):
        nx_detector.y_flipped = 12

    nx_detector.x_flipped = True
    nx_detector.y_flipped = False

    # check detector distance
    with pytest.raises(TypeError):
        nx_detector.distance = "test"
    nx_detector.distance = 0.02

    # check field of view
    with pytest.raises(ValueError):
        nx_detector.field_of_view = "test"
    nx_detector.field_of_view = FieldOfView.HALF

    # check count time
    with pytest.raises(TypeError):
        nx_detector.count_time = 12
    nx_detector.count_time = [0.1] * 5

    # check estimated cor from motor
    with pytest.raises(TypeError):
        nx_detector.estimated_cor_from_motor = "test"
    nx_detector.estimated_cor_from_motor = 0.5

    assert isinstance(nx_detector.to_nx_dict(), dict)

    # check we can't set undefined attributes
    with pytest.raises(AttributeError):
        nx_detector.test = 12

    # test nx_detector concatenation
    concatenated_nx_detector = NXdetector.concatenate([nx_detector, nx_detector])
    numpy.testing.assert_array_equal(
        concatenated_nx_detector.image_key_control, [ImageKey.PROJECTION] * 10
    )
    assert concatenated_nx_detector.x_pixel_size.value == 1e-7
    assert concatenated_nx_detector.y_pixel_size.value == 2e-7
    assert concatenated_nx_detector.distance.value == 0.02
    nx_detector.field_of_view = FieldOfView.HALF
    nx_detector.count_time = [0.1] * 10
    nx_detector.estimated_cor_from_motor = 0.5


def test_nx_detector_with_unit():
    diode = NXdetectorWithUnit(
        node_name="diode",
        expected_dim=(1,),
        default_unit=VoltageSystem.VOLT,
    )
    with pytest.raises(ValueError):
        diode.data = numpy.arange(10 * 10).reshape([10, 10])
    with pytest.raises(TypeError):
        diode.data = [10, 12]
    with pytest.raises(TypeError):
        diode.data = "test"
    diode.data = None
    diode.data = numpy.random.random(12)
    diode.data = (DataUrl(),)

    # test nx_detector concatenation
    concatenated_nx_detector = NXdetectorWithUnit.concatenate(
        [diode, diode],
        expected_dim=(1,),
        default_unit=VoltageSystem.VOLT,
    )
    assert len(concatenated_nx_detector.data.value) == 2
    assert isinstance(concatenated_nx_detector.data.value[1], DataUrl)


def test_nx_detector_with_virtual_source():
    """Insure detector data can be write from Virtual sources"""
    cwd = os.getcwd()
    with tempfile.TemporaryDirectory() as tmp_folder:
        # create virtual dataset
        n_base_raw_dataset = 5
        n_z, n_y, n_x = 4, 100, 100
        base_raw_dataset_shape = (n_z, n_y, n_x)
        n_base_raw_dataset_elmts = n_z * n_y * n_x

        v_sources = []

        raw_files = [
            os.path.join(tmp_folder, f"raw_file_{i_file}.hdf5")
            for i_file in range(n_base_raw_dataset)
        ]
        for i_raw_file, raw_file in enumerate(raw_files):
            with h5py.File(raw_file, mode="w") as h5f:
                h5f["data"] = numpy.arange(
                    start=n_base_raw_dataset_elmts * i_raw_file,
                    stop=n_base_raw_dataset_elmts * (i_raw_file + 1),
                ).reshape(base_raw_dataset_shape)
                v_sources.append(h5py.VirtualSource(h5f["data"]))

        nx_detector = NXdetector()
        nx_detector.data = v_sources

        detector_file = os.path.join(tmp_folder, "detector_file.hdf5")
        nx_detector.save(file_path=detector_file)

        # check the virtual dataset has been properly createde and linked
        with h5py.File(detector_file, mode="r") as h5f_master:
            dataset = h5f_master["/detector/data"]
            assert dataset.is_virtual
            for i_raw_file, raw_file in enumerate(raw_files):
                with h5py.File(raw_file, mode="r") as h5f_raw:
                    numpy.testing.assert_array_equal(
                        dataset[i_raw_file * n_z : (i_raw_file + 1) * n_z],
                        h5f_raw["data"],
                    )
            # check attributes have beem rewrite as expected
            assert "interpretation" in dataset.attrs

            # check virtual dataset is composed of relative links
            for vs_info in dataset.virtual_sources():
                assert vs_info.file_name.startswith("./")
        assert cwd == os.getcwd()

    # check concatenation
    concatenated_nx_detector = NXdetector.concatenate([nx_detector, nx_detector])
    assert isinstance(concatenated_nx_detector.data[1], h5py.VirtualSource)
    assert len(concatenated_nx_detector.data) == len(raw_files) * 2


def test_nx_detector_with_local_urls():
    """Insure detector data can be write from DataUrl linking to local dataset (in the same file)"""

    cwd = os.getcwd()
    n_base_dataset = 3
    n_z, n_y, n_x = 2, 10, 20
    base_dataset_shape = (n_z, n_y, n_x)
    n_base_dataset_elmts = n_z * n_y * n_x
    urls = []

    with tempfile.TemporaryDirectory() as tmp_folder:
        master_file = os.path.join(tmp_folder, "master_file.hdf5")
        with h5py.File(master_file, mode="a") as h5f:
            for i in range(n_base_dataset):
                data_path = f"/data_{i}"
                h5f[data_path] = numpy.arange(
                    start=n_base_dataset_elmts * i,
                    stop=n_base_dataset_elmts * (i + 1),
                ).reshape(base_dataset_shape)
                urls.append(
                    DataUrl(
                        file_path=master_file,
                        data_path=data_path,
                        scheme="silx",
                    )
                )
        nx_detector = NXdetector()
        nx_detector.data = urls
        nx_detector.save(file_path=master_file)

        # check the virtual dataset has been properly createde and linked
        with h5py.File(master_file, mode="r") as h5f_master:
            dataset = h5f_master["/detector/data"]
            assert dataset.is_virtual
            for i in range(n_base_dataset):
                numpy.testing.assert_array_equal(
                    dataset[i * n_z : (i + 1) * n_z],
                    numpy.arange(
                        start=n_base_dataset_elmts * i,
                        stop=n_base_dataset_elmts * (i + 1),
                    ).reshape(base_dataset_shape),
                )
            # check virtual dataset is composed of relative links
            for vs_info in dataset.virtual_sources():
                assert vs_info.file_name.startswith("./")
        assert cwd == os.getcwd()

    # check concatenation
    concatenated_nx_detector = NXdetector.concatenate([nx_detector, nx_detector])
    assert isinstance(concatenated_nx_detector.data[1], DataUrl)
    assert len(concatenated_nx_detector.data) == n_base_dataset * 2


def test_nx_detector_with_external_urls():
    """Insure detector data can be write from DataUrl linking to external dataset"""
    cwd = os.getcwd()
    with tempfile.TemporaryDirectory() as tmp_folder:
        # create virtual dataset
        n_base_raw_dataset = 5
        n_z, n_y, n_x = 4, 100, 100
        base_raw_dataset_shape = (n_z, n_y, n_x)
        n_base_raw_dataset_elmts = n_z * n_y * n_x

        urls = []

        raw_files = [
            os.path.join(tmp_folder, f"raw_file_{i_file}.hdf5")
            for i_file in range(n_base_raw_dataset)
        ]
        for i_raw_file, raw_file in enumerate(raw_files):
            with h5py.File(raw_file, mode="w") as h5f:
                h5f["data"] = numpy.arange(
                    start=n_base_raw_dataset_elmts * i_raw_file,
                    stop=n_base_raw_dataset_elmts * (i_raw_file + 1),
                ).reshape(base_raw_dataset_shape)
                # provide one file path each two as an absolue path
                if i_raw_file % 2 == 0:
                    file_path = os.path.abspath(raw_file)
                else:
                    file_path = os.path.relpath(raw_file, tmp_folder)
                urls.append(
                    DataUrl(
                        file_path=file_path,
                        data_path="data",
                        scheme="silx",
                    )
                )

        nx_detector = NXdetector()
        nx_detector.data = urls

        detector_file = os.path.join(tmp_folder, "detector_file.hdf5")
        # needed as we provide some link with relative path
        with cwd_context(tmp_folder):
            nx_detector.save(file_path=detector_file)

        # check the virtual dataset has been properly createde and linked
        with h5py.File(detector_file, mode="r") as h5f_master:
            dataset = h5f_master["/detector/data"]
            assert dataset.is_virtual
            for i_raw_file, raw_file in enumerate(raw_files):
                with h5py.File(raw_file, mode="r") as h5f_raw:
                    numpy.testing.assert_array_equal(
                        dataset[i_raw_file * n_z : (i_raw_file + 1) * n_z],
                        h5f_raw["data"],
                    )
            # check virtual dataset is composed of relative links
            for vs_info in dataset.virtual_sources():
                assert vs_info.file_name.startswith("./")

        assert cwd == os.getcwd()
        # check concatenation
        concatenated_nx_detector = NXdetector.concatenate([nx_detector, nx_detector])
        assert isinstance(concatenated_nx_detector.data[1], DataUrl)
        assert len(concatenated_nx_detector.data) == n_base_raw_dataset * 2


@pytest.mark.parametrize(
    "load_data_as, expected_type",
    [
        ("as_numpy_array", numpy.ndarray),
        ("as_virtual_source", h5py.VirtualSource),
        ("as_data_url", DataUrl),
    ],
)
def test_load_detector_data(tmp_path, load_data_as, expected_type):
    layout = h5py.VirtualLayout(shape=(4 * 2, 100, 100), dtype="i4")

    for n in range(0, 4):
        filename = os.path.join(tmp_path, "{n}.h5")
        with h5py.File(filename, "w") as f:
            f["data"] = numpy.arange(100 * 100 * 2).reshape(2, 100, 100)

        vsource = h5py.VirtualSource(filename, "data", shape=(2, 100, 100))
        start_n = n * 2
        end_n = start_n + 2
        layout[start_n:end_n] = vsource

    output_file = os.path.join(tmp_path, "VDS.h5")
    with h5py.File(output_file, "w") as f:
        f.create_virtual_dataset("data", layout, fillvalue=-5)

    nx_detector = NXdetector()
    nx_detector._load(
        file_path=output_file,
        data_path="/",
        load_data_as=load_data_as,
        nexus_version=None,
    )

    if expected_type is numpy.ndarray:
        assert isinstance(nx_detector.data, expected_type)
    else:
        for elmt in nx_detector.data:
            assert isinstance(elmt, expected_type)
    nx_detector.save(os.path.join(tmp_path, "output_file.nx"))


def test_nxtransformations_with_nxdetector(tmp_path):
    """
    test behavior of an Nxtransformations with an NXtomo
    and coherence between lr_flip / ud_flip API (provided for convenience) and providing directly the
    transformations
    """

    def build_detector():
        nx_detector = NXdetector(expected_dim=(2, 3))
        nx_detector.data = numpy.random.random(100 * 100 * 5).reshape(5, 100, 100)
        nx_detector.image_key_control = [1] * 5
        nx_detector.image_key_control = [ImageKey.PROJECTION] * 5
        return nx_detector

    nx_detector_1 = build_detector()
    nx_detector_2 = build_detector()

    # test having a left-right flip
    nx_detector_1.transformations.addTransformation(
        Transformation(
            axis_name="rz",
            value=180,
            transformation_type="rotation",
            vector=TransformationAxis.AXIS_Z,
        )
    )
    nx_detector_2.x_flipped = True

    assert (
        nx_detector_1.transformations.to_nx_dict()
        == nx_detector_2.transformations.to_nx_dict()
    )

    # test having a up-down flip
    nx_detector_3 = build_detector()
    nx_detector_4 = build_detector()
    nx_detector_3.transformations.addTransformation(
        Transformation(
            axis_name="ry",
            value=180,
            transformation_type="rotation",
            vector=TransformationAxis.AXIS_Y,
        )
    )
    nx_detector_4.y_flipped = True

    assert (
        nx_detector_3.transformations.to_nx_dict()
        == nx_detector_4.transformations.to_nx_dict()
    )

    # having both lr and ud
    nx_detector_5 = build_detector()
    nx_detector_6 = build_detector()

    nx_detector_5.transformations.addTransformation(
        Transformation(
            axis_name="ry",
            value=180,
            transformation_type="rotation",
            vector=TransformationAxis.AXIS_Y,
        )
    )
    nx_detector_5.transformations.addTransformation(
        Transformation(
            axis_name="rz",
            value=180,
            transformation_type="rotation",
            vector=TransformationAxis.AXIS_Z,
        )
    )
    nx_detector_6.x_flipped = True
    nx_detector_6.y_flipped = True

    assert (
        nx_detector_5.transformations.to_nx_dict()
        == nx_detector_6.transformations.to_nx_dict()
    )



import os
from datetime import datetime
from tempfile import TemporaryDirectory
import numpy

import pytest
from silx.io.url import DataUrl

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import FieldOfView, ImageKey
from nxtomo.nxobject.utils import concatenate


nexus_path_versions = (1.1, 1.0, None)


@pytest.mark.parametrize("nexus_path_version", nexus_path_versions)
def test_nx_tomo(nexus_path_version):
    nx_tomo = NXtomo(node_name="")

    # check start time
    with pytest.raises(TypeError):
        nx_tomo.start_time = 12
    nx_tomo.start_time = datetime.now()

    # check end time
    with pytest.raises(TypeError):
        nx_tomo.end_time = 12
    nx_tomo.end_time = datetime(2022, 2, 27)

    # check sample
    with pytest.raises(TypeError):
        nx_tomo.sample = "tata"

    # check detector
    with pytest.raises(TypeError):
        nx_tomo.instrument.detector = "tata"

    # check energy
    with pytest.raises(TypeError):
        nx_tomo.energy = "tata"
    nx_tomo.energy = 12.3

    # check group size
    with pytest.raises(TypeError):
        nx_tomo.group_size = "tata"
    nx_tomo.group_size = 3

    # check title
    with pytest.raises(TypeError):
        nx_tomo.title = 12
    nx_tomo.title = "title"

    # check instrument
    with pytest.raises(TypeError):
        nx_tomo.instrument = "test"

    # check we can't set undefined attributes
    with pytest.raises(AttributeError):
        nx_tomo.test = 12

    # create detector for test
    projections = numpy.random.random(100 * 100 * 8).reshape([8, 100, 100])
    flats_1 = numpy.random.random(100 * 100 * 2).reshape([2, 100, 100])
    darks = numpy.random.random(100 * 100 * 3).reshape([3, 100, 100])
    flats_2 = numpy.random.random(100 * 100 * 2).reshape([2, 100, 100])
    alignment = numpy.random.random(100 * 100 * 1).reshape([1, 100, 100])
    nx_tomo.instrument.detector.data = numpy.concatenate(
        [
            darks,
            flats_1,
            projections,
            flats_2,
            alignment,
        ]
    )

    nx_tomo.instrument.detector.image_key_control = numpy.concatenate(
        [
            [ImageKey.DARK_FIELD] * 3,
            [ImageKey.FLAT_FIELD] * 2,
            [ImageKey.PROJECTION] * 8,
            [ImageKey.FLAT_FIELD] * 2,
            [ImageKey.ALIGNMENT] * 1,
        ]
    )

    nx_tomo.instrument.detector.x_pixel_size = (
        nx_tomo.instrument.detector.y_pixel_size
    ) = 1e-7
    nx_tomo.instrument.detector.distance = 0.2
    nx_tomo.instrument.detector.field_of_view = FieldOfView.HALF
    nx_tomo.instrument.detector.count_time = numpy.concatenate(
        [
            [0.2] * 3,  # darks
            [0.1] * 2,  # flats 1
            [0.1] * 8,  # projections
            [0.1] * 2,  # flats 2
            [0.1] * 1,  # alignment
        ]
    )

    # create sample for test
    nx_tomo.sample.name = "my sample"
    nx_tomo.sample.rotation_angle = numpy.concatenate(
        [
            [0.0] * 3,  # darks
            [0.0] * 2,  # flats 1
            numpy.linspace(0, 180, num=8, endpoint=False),  # projections
            [180.0] * 2,  # flats 2
            [0.0],  # alignement
        ]
    )

    n_frames = 3 + 2 + 8 + 2 + 1
    nx_tomo.sample.x_translation = [0.6] * n_frames
    nx_tomo.sample.y_translation = [0.2] * n_frames
    nx_tomo.sample.z_translation = [0.1] * n_frames
    assert nx_tomo.is_root is True
    assert nx_tomo.instrument.is_root is False
    assert (
        nx_tomo.root_path
        == nx_tomo.instrument.root_path
        == nx_tomo.instrument.detector.root_path
    )

    NXtomo.check_consistency(nx_tomo=nx_tomo, raises_error=True)

    with TemporaryDirectory() as folder:
        file_path = os.path.join(folder, "nexus_file.hdf5")

        nx_tomo.save(
            file_path=file_path,
            data_path="entry",
            nexus_path_version=nexus_path_version,
        )
        assert os.path.exists(file_path)

        # insure we can read it back
        scan = NXtomo().load(file_path, data_path="entry")
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.FLAT_FIELD,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 4
        )
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.DARK_FIELD,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 3
        )
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.PROJECTION,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 8
        )
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.ALIGNMENT,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 1
        )
        assert scan.energy.value == 12.3
        assert scan.instrument.detector.x_pixel_size.value == 1e-7
        assert scan.instrument.detector.y_pixel_size.value == 1e-7
        assert scan.instrument.detector.distance.value == 0.2
        assert scan.instrument.detector.field_of_view == FieldOfView.HALF
        assert scan.sample.name == "my sample"
        assert (
            len(scan.sample.x_translation.value)
            == len(scan.sample.y_translation.value)
            == len(scan.sample.z_translation.value)
            == n_frames
        )
        assert scan.sample.x_translation.value[0] == 0.6
        assert scan.sample.y_translation.value[0] == 0.2
        assert scan.sample.z_translation.value[0] == 0.1
        if nexus_path_version != 1.0:
            assert scan.instrument.source.name is not None
            assert scan.instrument.source.type is not None

        # try to load it from the disk
        loaded_nx_tomo = NXtomo("test").load(file_path=file_path, data_path="entry")
        assert isinstance(loaded_nx_tomo, NXtomo)
        assert loaded_nx_tomo.energy.value == nx_tomo.energy.value
        assert loaded_nx_tomo.energy.unit == nx_tomo.energy.unit
        assert loaded_nx_tomo.start_time == nx_tomo.start_time
        assert loaded_nx_tomo.end_time == nx_tomo.end_time
        assert (
            loaded_nx_tomo.instrument.detector.x_pixel_size.value
            == nx_tomo.instrument.detector.x_pixel_size.value
        )
        assert (
            loaded_nx_tomo.instrument.detector.x_pixel_size.unit
            == nx_tomo.instrument.detector.x_pixel_size.unit
        )
        assert (
            loaded_nx_tomo.instrument.detector.y_pixel_size.value
            == nx_tomo.instrument.detector.y_pixel_size.value
        )
        assert (
            loaded_nx_tomo.instrument.detector.field_of_view
            == nx_tomo.instrument.detector.field_of_view
        )
        numpy.testing.assert_array_equal(
            loaded_nx_tomo.instrument.detector.count_time.value,
            nx_tomo.instrument.detector.count_time.value,
        )
        assert (
            loaded_nx_tomo.instrument.detector.count_time.unit
            == nx_tomo.instrument.detector.count_time.unit
        )
        assert (
            loaded_nx_tomo.instrument.detector.distance.value
            == nx_tomo.instrument.detector.distance.value
        )
        assert (
            loaded_nx_tomo.instrument.detector.distance.unit
            == nx_tomo.instrument.detector.distance.unit
        )
        assert (
            loaded_nx_tomo.instrument.detector.estimated_cor_from_motor
            == nx_tomo.instrument.detector.estimated_cor_from_motor
        )

        numpy.testing.assert_array_equal(
            loaded_nx_tomo.instrument.detector.image_key_control,
            nx_tomo.instrument.detector.image_key_control,
        )
        numpy.testing.assert_array_equal(
            loaded_nx_tomo.instrument.detector.image_key,
            nx_tomo.instrument.detector.image_key,
        )

        assert loaded_nx_tomo.sample.name == nx_tomo.sample.name
        assert loaded_nx_tomo.sample.rotation_angle is not None
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.rotation_angle, nx_tomo.sample.rotation_angle
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.x_translation.value,
            nx_tomo.sample.x_translation.value,
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.y_translation.value,
            nx_tomo.sample.y_translation.value,
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.z_translation.value,
            nx_tomo.sample.z_translation.value,
        )

        loaded_nx_tomo = NXtomo("test").load(
            file_path=file_path, data_path="entry", detector_data_as="as_numpy_array"
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.instrument.detector.data,
            nx_tomo.instrument.detector.data,
        )
        loaded_nx_tomo = NXtomo("test").load(
            file_path=file_path, data_path="entry", detector_data_as="as_data_url"
        )
        assert isinstance(loaded_nx_tomo.instrument.detector.data[0], DataUrl)
        with pytest.raises(ValueError):
            # check an error is raise because the dataset is not virtual
            loaded_nx_tomo = NXtomo("test").load(
                file_path=file_path,
                data_path="entry",
                detector_data_as="as_virtual_source",
            )

        # test concatenation
        nx_tomo_concat = concatenate([loaded_nx_tomo, None, loaded_nx_tomo])
        concat_file = os.path.join(folder, "concatenated_nexus_file.hdf5")

        nx_tomo_concat.save(
            file_path=concat_file,
            data_path="myentry",
            nexus_path_version=nexus_path_version,
        )
        loaded_concatenated_nx_tomo = NXtomo("test").load(
            file_path=concat_file,
            data_path="myentry",
            detector_data_as="as_virtual_source",
        )
        numpy.testing.assert_array_almost_equal(
            loaded_concatenated_nx_tomo.sample.rotation_angle,
            numpy.concatenate(
                [
                    nx_tomo.sample.rotation_angle,
                    nx_tomo.sample.rotation_angle,
                ]
            ),
        )
        numpy.testing.assert_array_almost_equal(
            loaded_concatenated_nx_tomo.sample.x_translation.value,
            numpy.concatenate(
                [
                    nx_tomo.sample.x_translation.value,
                    nx_tomo.sample.x_translation.value,
                ]
            ),
        )

        with pytest.raises(TypeError):
            concatenate([1, 2])


@pytest.mark.parametrize("nexus_path_version", nexus_path_versions)
def test_nx_tomo_subselection(nexus_path_version):
    """
    test sub_select_from_projection_angle_range
    """
    nx_tomo = NXtomo()
    nx_tomo.energy = 12.3
    shape = (12, 12)
    data_dark = numpy.ones(shape)
    data_flat = numpy.ones(shape) * 2.0
    data_projection = numpy.ones(shape) * 3.0
    str(nx_tomo)
    nx_tomo.instrument.detector.data = numpy.concatenate(
        (
            data_dark,
            data_dark,
            data_flat,
            data_projection,
            data_projection,
            data_projection,
            data_flat,
            data_projection,
            data_projection,
            data_projection,
            data_flat,
        )
    )
    nx_tomo.instrument.detector.image_key_control = numpy.array(
        (
            ImageKey.DARK_FIELD,
            ImageKey.DARK_FIELD,
            ImageKey.FLAT_FIELD,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.FLAT_FIELD,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.FLAT_FIELD,
        )
    )

    original_angles = numpy.array(
        (
            0,
            0,
            0,
            10,
            20.5,
            22.5,
            180,
            180,
            200,
            300.2,
            300.2,
        )
    )
    nx_tomo.sample.rotation_angle = original_angles

    nx_tomo_sub_1 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=10,
        angle_interval=12.5,
        shift_angles=False,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_1.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.INVALID,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.INVALID,
                ImageKey.INVALID,
                ImageKey.INVALID,
                ImageKey.FLAT_FIELD,
            )
        ),
    )
    numpy.testing.assert_equal(
        nx_tomo_sub_1.sample.rotation_angle,
        original_angles,
    )

    nx_tomo_sub_2 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=10,
        angle_interval=20,
        shift_angles=True,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_2.sample.rotation_angle[0:3],
        0.0,
    )
    numpy.testing.assert_array_equal(
        nx_tomo_sub_2.sample.rotation_angle[3:6],
        numpy.array([0.0, 10.5, 12.5]),
    )

    nx_tomo_sub_3 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=-10,
        angle_interval=300,
        shift_angles=False,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_3.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.INVALID,
                ImageKey.FLAT_FIELD,
            )
        ),
    )

    nx_tomo_sub_4 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=-10,
        angle_interval=None,
        shift_angles=False,
    )
    numpy.testing.assert_equal(
        nx_tomo_sub_4.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
            )
        ),
    )


def test_bliss_original_files(tmp_path):
    """
    test about NXtomo.bliss_original_files
    """
    test_dir = tmp_path / "test_bliss_original_files"
    test_dir.mkdir()

    nx_tomo_1 = NXtomo()
    with pytest.raises(TypeError):
        nx_tomo_1.bliss_original_files = 12

    nx_tomo_1.bliss_original_files = ("/path/1", "/path/2")

    nx_tomo_2 = NXtomo()
    nx_tomo_2.bliss_original_files = ("/path/2", "/path/3")

    nx_tomo_3 = NXtomo()

    nx_tomo_4 = NXtomo()
    nx_tomo_4.bliss_original_files = ()

    nx_tomo_concat = concatenate([nx_tomo_1, nx_tomo_2, nx_tomo_3])
    assert nx_tomo_concat.bliss_original_files == ("/path/1", "/path/2", "/path/3")

    output_nx_tomo_concat = os.path.join(test_dir, "nx_concat.nx")
    nx_tomo_concat.save(output_nx_tomo_concat, "/entry_concat")

    loaded_nx_tomo = NXtomo().load(output_nx_tomo_concat, "/entry_concat")
    assert loaded_nx_tomo.bliss_original_files == ("/path/1", "/path/2", "/path/3")

    output_nx_tomo_file = os.path.join(test_dir, "nx_tomo.nx")
    nx_tomo_3.save(output_nx_tomo_file, "/entry0000")
    loaded_nx_tomo = NXtomo().load(output_nx_tomo_file, "/entry0000")
    assert loaded_nx_tomo.bliss_original_files is None

    nx_tomo_4.save(output_nx_tomo_file, "/entry0000", overwrite=True)
    loaded_nx_tomo = NXtomo().load(output_nx_tomo_file, "/entry0000")
    assert loaded_nx_tomo.bliss_original_files == ()
